(*  Title:      Provers/Arith/cancel_sums.ML
    ID:         $Id: cancel_sums.ML,v 1.3 1997/12/19 09:34:27 wenzelm Exp $
    Author:     Markus Wenzel and Stefan Berghofer, TU Muenchen

Modified by Jeremy Dawson to allow for variable types (noted by VT or Vt)

Cancel common summands of balanced expressions:

  A + x + B ~~ A' + x + B'  ==  A + B ~~ A' + B'

where + is AC0 and ~~ an appropriate balancing operation (e.g. =, <=, <, -).
*)

fun upf _ = use "pmg_cancel_sums.ML" ;

signature VT_CANCEL_SUMS_DATA =
sig
  (*abstract syntax*)
  val mk_sum: typ -> term list -> term (* typ argument added - VT *)
  val dest_sum: term -> term list
  val mk_bal: term * term -> term
  val dest_bal: term -> term * term
  (*rules*)
  val prove_conv: tactic -> tactic -> Sign.sg -> term * term -> thm
  val norm_tac: tactic                   	(*AC0 etc.*)
  val uncancel_tac: cterm -> tactic    	(*apply A ~~ B  ==  x + A ~~ x + B*)
end;

signature VT_CANCEL_SUMS =
sig
  val proc: Sign.sg -> simpset -> term -> thm option
end;

functor VtCancelSumsFun(Data: VT_CANCEL_SUMS_DATA): VT_CANCEL_SUMS =
struct

(* cancel *)

fun cons1 x (xs, y, z) = (x :: xs, y, z);
fun cons2 y (x, ys, z) = (x, y :: ys, z);
fun cons12 x y (xs, ys, z) = (x :: xs, y :: ys, z);

(*symmetric difference of multisets --
  assumed to be sorted wrt. Logic.term_ord*)
fun cancel ts [] vs = (ts, [], vs)
  | cancel [] us vs = ([], us, vs)
  | cancel (t :: ts) (u :: us) vs =
      (case Term.term_ord (t, u) of
        EQUAL => cancel ts us (t :: vs)
      | LESS => cons1 t (cancel ts (u :: us) vs)
      | GREATER => cons2 u (cancel (t :: ts) us vs));

(* uncancel *)

fun uncancel_sums_tac _ [] = all_tac
  | uncancel_sums_tac sg (t :: ts) =
      Data.uncancel_tac (Thm.cterm_of sg t) THEN
      uncancel_sums_tac sg ts;

(* the simplification procedure *)

fun proc sg _ t =
  (case try Data.dest_bal t of
    NONE => NONE
  | SOME (bal as (bl, _)) =>
      let
        val T = Term.fastype_of bl ; (* VT *)
        val (ts, us) = pairself (sort Term.term_ord o Data.dest_sum) bal;
        val (ts', us', vs) = cancel ts us [];
      in
        if null vs then NONE
        else SOME
          (Data.prove_conv (uncancel_sums_tac sg vs) Data.norm_tac sg
            (t, Data.mk_bal (Data.mk_sum T ts', Data.mk_sum T us')))
      end);

end; (* functor VtCancelSumsFun *)

(** actual simplification procedures for class pm_ge0 **)
(* refer source files src/Provers/Arith/cancel_sums.ML
  examples in ArithData.nat_cancel_sums in src/HOL/arith_data.ML *)

(* mk_sum, mk_norm_sum *)

val mk_plus = HOLogic.mk_binop "op +";

(* VT - changed from original, to omit 0 if possible *)
(* also typ argument added - VT *)
fun mk_sum T []        = Const ("0", T) 
  | mk_sum T [t]       = t (* replaced [t, u] VT *)
  | mk_sum T (t :: ts) = mk_plus (t, mk_sum T ts);

(* dest_sum *)

val dest_plus = HOLogic.dest_bin "op +" Term.dummyT ;

fun dest_sum tm =
  if HOLogic.is_zero tm then []
  else
    (case try dest_plus tm of
      SOME (t, u) => dest_sum t @ dest_sum u
    | NONE => [tm]);

(** generic proof tools **)

(* prove conversions *)

val mk_eqv = HOLogic.mk_Trueprop o HOLogic.mk_eq;

fun prove_conv expand_tac norm_tac sg tu =
  mk_meta_eq (prove_goalw_cterm_nocheck [] (cterm_of sg (mk_eqv tu))
    (K [expand_tac, norm_tac]))
  handle ERROR => error ("The error(s) above occurred while trying to prove " ^
    (string_of_cterm (cterm_of sg (mk_eqv tu))));

val subst_equals = prove_goal HOL.thy "[| t = s; u = t |] ==> u = s"
  (fn prems => [cut_facts_tac prems 1, SIMPSET' asm_simp_tac 1]);

(* rewriting *)

fun simp_all rules = ALLGOALS (simp_tac (HOL_ss addsimps rules));

val add_rules = [plus_ac0_zero, plus_ac0_zero_right];

fun prep_simproc (name, pats, proc) =
  Simplifier.simproc (Theory.sign_of (the_context ())) name pats proc;

(*
set show_sorts ;
reset show_sorts ;
set trace_simp ;
reset trace_simp ;
*)

(** cancel common summands **)

structure Sum =
struct
  val mk_sum = mk_sum;
  val dest_sum = dest_sum;
  val prove_conv = prove_conv;
  val norm_tac = simp_all add_rules THEN simp_all plus_ac;
end;

exception PMGI of (cterm * thm) ;

fun gen_uncancel_tac rule ct =
  rtac (instantiate' [SOME (Thm.ctyp_of_term ct)] [NONE, SOME ct] 
    (rule RS subst_equals)) 1
  handle _ => raise PMGI (ct, rule) ;

structure PmgEqCancelSums = VtCancelSumsFun
(struct open Sum;
  val mk_bal = HOLogic.mk_eq;
  val dest_bal = HOLogic.dest_bin "op =" Term.dummyT;
  val uncancel_tac = gen_uncancel_tac add_left_canc_eq ;
end);

structure PmgLessCancelSums = VtCancelSumsFun
(struct open Sum;
  val mk_bal = HOLogic.mk_binrel "op <";
  val dest_bal = HOLogic.dest_bin "op <" Term.dummyT;
  val uncancel_tac = gen_uncancel_tac add_left_canc_less ;
end);

structure PmgLeCancelSums = VtCancelSumsFun
(struct open Sum;
  val mk_bal = HOLogic.mk_binrel "op <=";
  val dest_bal = HOLogic.dest_bin "op <=" Term.dummyT;
  val uncancel_tac = gen_uncancel_tac add_left_canc_le ;
end);

structure PmgDiffCancelSums = VtCancelSumsFun
(struct open Sum;
  val mk_bal = HOLogic.mk_binop "op -";
  val dest_bal = HOLogic.dest_bin "op -" Term.dummyT;
  val uncancel_tac = gen_uncancel_tac add_left_canc_diff ;
end);

(** prepare pmg_cancel simprocs **)

val pmg_cancel_sums_add = map prep_simproc
  [("pmgeq_cancel_sums",
     ["(l::'a::pm0) + m = n", "(l::'a::pm0) = m + n"],
     PmgEqCancelSums.proc),
   ("pmgless_cancel_sums",
     ["(l::'a::pm_ge0) + m < n", "(l::'a::pm_ge0) < m + n"],
     PmgLessCancelSums.proc),
   ("pmgle_cancel_sums",
     ["(l::'a::pm_ge0) + m <= n", "(l::'a::pm_ge0) <= m + n"],
     PmgLeCancelSums.proc)];

val pmg_cancel_sums as pmgsps = pmg_cancel_sums_add @
  [prep_simproc ("pmgdiff_cancel_sums",
    ["((l::'a::pm0) + m) - n", "(l::'a::pm0) - (m + n)"],
    PmgDiffCancelSums.proc)];

(* 
Goal "X" ;
val thy = the_context () ;
val sg = sign_of thy ;
(* fn_of : Simplifier.simproc -> term -> thm option *)
fun fn_of simproc = (#3 o FullSimplifier.rep_simproc) simproc sg [] ;
fun name_of simproc = (#1 o FullSimplifier.rep_simproc) simproc ;

val _ = pmg_cancel_sums : Simplifier.simproc list ;
map FullSimplifier.rep_simproc pmg_cancel_sums ;

val ncsfs as [feq, fless, fle, fdiff] = map fn_of pmg_cancel_sums ;
val names = map name_of pmg_cancel_sums ;
val ts = map read ["(a :: 'a :: pm_ge0) + b = a + c",
  "e + (a :: 'a :: pm0) + b = d + a",
  "(a :: nat) + b = d + a + c",
  "e + (a :: 'a :: abelian_group) + b = d + a + c"] ;
val theqs = map feq ts ;

val ts = map read [
      "e + (a :: int) + b = d + a + c",
      "e + (a :: nat) + b < d + a + c",
      "e + (a :: nat) + b <= d + a + c",
      "(e + (a :: 'a :: abelian_group) + b) - (d + a + c)" ] ;
val ths = ListPair.map (uncurry I) (ncsfs, ts) ;
*)
